/*
 * Usage: makedatafiles -r <randomness> -s <size> -o <outputfile>
 *
 * Makes a data file in <outputfile> of size <size> bytes.  The amount of randomness
 * is determined by the parameter <randomness>.
 *
 * Randomness is a float between 0 and 1; 0 = completely non-random, 1 = completely
 * random data.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define DATA_FILE_MAX 128
#define ERR_MSG_SIZE 128
#define SEED 1235791113
#define MAX_UNSIGNED_INT ( ~ ( (unsigned int) 0) )


void usage()
{
  printf ("Usage: makedatafiles -r <0.0 .. 0.1> -s <size> -o <outputfile>\n");
  exit (1);
}

/* if next opt begins with '-', return following char, else usage() */
char getnextopt(char *argv[])
{
  if ((*argv)[0] != '-')    
    usage();

  return ((*argv)[1]);
}  


void parse_args (int argc, char *argv[], float *randomness, char data_file[], int *size)
{

  if (argc != 7) {
    usage();
    exit (-1);
  }

  argv++;

  while (*argv) {
    
    switch (getnextopt(argv)) {

    case 'r':
      argv++;
      sscanf(*argv++, "%f", randomness);
      if ( (*randomness < 0.0) || (*randomness > 1.0) ) {
	printf ("Randomness parameter must be between 0.0 and 1.0.\n");
	usage();
	exit (-1);
      }
      argc--;
      break;
      
    case 'o':
      argv++;
      strncpy(data_file, *argv++, DATA_FILE_MAX);
      argc--;
      break;
      
    case 's':
      argv++;
      *size = atoi(*argv++);
      argc--;
      break;
      
    default:
      usage();
      break;
    }
  }
}


/* Determine whether or not the data pattern being written to
 * the file should be perturbed.
 */
int disturb (float r)
{

  return ( (float) rand() / (float) RAND_MAX < r );

}

  
/* Create the random data file.  The randomness
 * parameter dictates how often the byte patterns written to 
 * the file change.  
 *
 * The random file rfd is the source of both random byte patterns
 * and irregularity within the sequence of bytes.
 */
void create_file (FILE *dfd, float r, int size)
{

  char c;

  /* Sanity check */
  if (size < 0) {
    printf ("Requested file size < 0.\n");
    exit (-1);
  }

  c = (unsigned char) 0;

  while (size) {

    if (disturb (r))
      c = (unsigned char) rand();

    if (fputc ((int) c, dfd) == EOF) {
      perror ("fputc");
      exit (-1);
    }

    size--;

  }

}
    

int main (int argc, char* argv[])
{
  float randomness;
  char data_file[DATA_FILE_MAX];
  int size;

  FILE *dfd;
  char err_msg[ERR_MSG_SIZE];

  parse_args (argc, argv, &randomness, data_file, &size);

  srand(SEED);

  if ( (dfd = fopen (data_file, "w")) < 0) {
    sprintf (err_msg, "Can't open %s", data_file);
    perror (err_msg);
    exit (-1);
  }

  create_file (dfd, randomness, size);

  if (fclose(dfd) < 0) {
    perror (data_file);
    exit (-1);
  }

  exit(0);
}
