#!/bin/bash
#
# A driver program for the FS test suite
#

FS_TEST_HOME_DIR=`pwd`/fs
TEMP_DIR="/tmp"

pushd "$FS_TEST_HOME_DIR"

#
# Holey file test
#
HOLEY_FILE="$TEMP_DIR"/holey-$$
./holey $HOLEY_FILE &
CHILD=$!

trap "kill -9 $CHILD ; rm $HOLEY_FILE >> /dev/null 2>&1 ; exit -1" SIGINT SIGHUP SIGTERM
wait

[ $? = 0 ] || { echo "Holey test failed with error code $?." ; exit -1 ; }
rm $HOLEY_FILE
[ $? ] || { echo "Couldn't remove file $HOLEY_FILE." ; exit -1 ; }

#
#directory test
#
./dir.sh "$TEMP_DIR" &

CHILD=$!

trap "kill -9 $CHILD ; rm -rf /tmp/dir{1,2,3,4,5,6,7,8,9}* >> /dev/null 2>&1 ; exit -1" SIGINT SIGHUP SIGTERM
wait $CHILD

[ $? = 0 ] || { echo "Dir test failed with error code $?." ; exit -1 ; }

#
#Sym links test
#
./slinks.sh "$TEMP_DIR" &
CHILD=$!

trap "kill -9 $CHILD ; rm -rf /tmp/slinks* >> /dev/null 2>&1 ; exit -1" SIGINT SIGHUP SIGTERM
wait $CHILD

[ $? = 0 ] || { echo "Sym links test failed with error code $?." ; exit -1 ; }

#
# SGI LTP doio test
#
TMP_FILE=ltp$$
LTP_FILE="$TEMP_DIR"/"$TMP_FILE"

#
# Change to ltp home dir
#
cd ../src/ltp/doio
[ $? = 0 ] || { echo "Can't cd to ltp home." ; exit -1 ; }

#
# Options used here
#
# iogen [options] [filesize[:filename]]
#   -i iterations
#   
# doio
#   -a abort; kill all doio processes on data err
#   -v verify writes
#   -n number of processes to start up
#   -U panic_conditions

trap "kill -9 `pidof iogen` `pidof doio` ; rm $LTP_FILE ; exit -1" SIGINT SIGHUP SIGTERM

./iogen -i 1500 1000000b:"$LTP_FILE" | ./doio -av -n 16 -U 'all'
[ $? = 0 ] || { echo "Doio failed with error number $?." ; exit -1 ; }

#
# Repeat the test with file locking.
# This should cause lots of nice juicy lock contention.
#
./iogen -i 1500 1000000b:"$LTP_FILE" | ./doio -avk -n 16 -U 'all'
[ $? = 0 ] || { echo "Doio with locking failed with error number $?." ; exit -1 ; }

rm "$LTP_FILE" > /dev/null 
[ $? = 0 ] || { echo "Error: can't remove file $FILE." ; exit -1 ; }

#
# Ingo's exorcist tests
#
# Usage: exorcist2.pl <nprocs> <ndirs> <nfiles> <filename> <iterations>
#
EX_FILE_BIG="$TEMP_DIR"/exbig
EX_FILE_SMALL="$TEMP_DIR"/exsmall

cd $TEMP_DIR

dd if=/dev/urandom of="$EX_FILE_BIG" bs=300000 count=1
[ $? = 0 ] || { echo 'Error: can not create $EX_FILE_BIG for exorcist.' ; exit -1 ; }

dd if=/dev/urandom of="$EX_FILE_SMALL" bs=1023 count=1
[ $? = 0 ] || { echo 'Error: can not create $EX_FILE_SMALL for exorcist.' ; exit -1 ; }

$FS_TEST_HOME_DIR/exorcist2.pl 16 32 64 $EX_FILE_SMALL 10 > /dev/null &
EX_PID_SMALL=$!

$FS_TEST_HOME_DIR/exorcist2.pl 16 32 64 $EX_FILE_BIG 10 > /dev/null &
EX_PID_BIG=$!

trap "kill -9 $EX_PID_SMALL $EX_PID_BIG ; rm -r $EX_FILE_SMALL $EX_FILE_BIG $TEMP_DIR/EXORCIST" SIGINT SIGTERM SIGHUP

#
# OK, Ingo says this sync loop will really drive the VFS nuts if executed
# while exorcist is running.  Why the 1..10 you ask?  Well, 100 seconds
# seemed to be just long enough in practice.  We'll see...

for SYNC_LOOP in 1 2 3 4 5 6 7 8 9 10 ; do {
    echo "Syncing... " `date`
    sync
    sleep 10
} ; done

sync

wait 

sync


rm -rf $EX_FILE_SMALL $EX_FILE_BIG $TEMP_DIR/EXORCIST >> /dev/null 2>&1


# Han's Reisers truncate test, modified to loop from file size down to zero.

TRUNCATE_FILE=$TEMP_DIR/truncate-$$
TRUNCATE_SIZE=2097152  # == 2MB

dd if=/dev/urandom of="$TRUNCATE_FILE" bs=1024 count=2048 >> /dev/null
[ $? = 0 ] || { echo "Error: can not create $FTRUNCATE_FILE for ftruncate test." ; exit -1 ; }

$FS_TEST_HOME_DIR/truncate $TRUNCATE_FILE $TRUNCATE_SIZE &

TRUNCATE_PID=$!
trap "kill -9 $TRUNCATE_PID ; rm -f $TRUNCATE_FILE" SIGINT SIGTERM SIGHUP

wait

[ $? = 0 ] || { echo "Truncate test returned $?" ; exit -1 ; }

rm -f $TRUNCATE_FILE >> /dev/null
[ $? = 0 ] || { echo "Can't remove $TRUNCATE_FILE." ; exit -1 ; }


# Ftruncate64 test: open64/ftruncate64/close/unlink.  Oops's 2.4.16
# Also appears to fail intermittently on 2.4.9.  This is known.

TMPFILE=`mktemp "$TEMP_DIR"/ftruncate64.XXXXXX`
[ $? = 0 ] || { echo "Can not make ftruncate64 temp file." ; exit -1 ; }

$FS_TEST_HOME_DIR/ftruncate64 $TMPFILE 1000 &
CHILD=$!
trap "kill -9 $CHILD ; rm $TMPFILE >> /dev/null 2>&1 ; exit -1" SIGINT SIGHUP SIGTERM

wait
[ $? = 0 ] || { echo "Ftruncate64 test failed." ; exit -1 ; }

exit 0




