/*
 * This test creates a file with a large hole in the middle and then
 * writes data in random locations in the file.  After the write, the
 * entire file is re-read and compared to a precomputed, in-memory 
 * image of the current contents of the file.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#define BLOCKSIZE 1024
#define NBLOCKS 128 * 128 + 128 + 1
#define FILE_SIZE BLOCKSIZE*NBLOCKS

#define MAX_FNAME 128
#define WORKING_DIR "/tmp/"
#define FILE_NAME_PREFIX "file_with_holes"

#define ITERATIONS 512
#define ERR_CONTEXT 5 /* Amount of context to print around an error */

#define SEED 123456789

/* 
 * Return a random int in the interval [low, high] 
 */
int randint (int low, int high)
{
  int normalized;

  normalized = (int) ( (float) rand() / 
		       ( (float) RAND_MAX / (float) (high - low + 1) ) );
  return (normalized + low);
}


int main (int argc, char *argv[])
{

  unsigned char *holey_image, *holey_image_mem, *buf;
  char fname[MAX_FNAME+8];
  char errmsg[100];
  int holey_fd;
  int start, size;
  int i, j;
  unsigned char c;
  int err;
  int newly_written, nwritten, buf_start, write_size;
  long long written;

  srand (SEED); /* init rand() */

  if (!argv[1]) {
    printf ("Usage: holey <fname>.\n");
    exit (-1);
  }
  strncpy (fname, argv[1], MAX_FNAME);

  if ((holey_fd = open (fname, O_RDWR | O_CREAT | O_TRUNC, 
			S_IRUSR | S_IWUSR)) < 0) {
    perror ("Can't open holey file %s.");
    exit (-1);
  }

  if ( (holey_image = (char *) malloc (FILE_SIZE)) == NULL) {
    printf ("Malloc of holey_image failed.\n");
    exit (-1);
  }

  if ( (holey_image_mem = (char *) malloc (FILE_SIZE)) == NULL) {
    printf ("Malloc of holey_image_mem failed.\n");
    exit (-1);
  }

  if ( (buf = (char *) malloc (FILE_SIZE)) == NULL) {
    printf ("Malloc of buf failed.\n");
    exit (-1);
  }

  c = (char) 0;
  /* initialize in-memory check array, and buffer */
  for (i = 0; i < FILE_SIZE; i++) {
    holey_image[i] = holey_image_mem[i] = (char) 0;
    buf[i] = c++;
  }

 /* make a file with a big hole at the beginning, initialize the 
  * in-memory image of the file
  */
  if (lseek (holey_fd, FILE_SIZE - 1, SEEK_SET) != FILE_SIZE - 1) {
    perror ("Initial write of holey file failed");
    exit (-1);
  }

  c = '1';
  if (write (holey_fd, &c, 1) != 1) {
    perror ("Write at and of hole failed");
    exit (-1);
  }
  holey_image[FILE_SIZE-1] = c;

  written = newly_written = 0;

  /* Ok, now let's start looping and writing data to random places in the
     file.  After each write, update the in-memory image of the file.
     Once that's done, re-read the file and compare to the in-mem image.
  */
  
  for (i = 0; i < ITERATIONS; i++) {
#ifdef DEBUG
    printf ("%d ", i);
    fflush(NULL);
#endif
    start = randint (0, FILE_SIZE - 1);
    size = randint (1, FILE_SIZE - start);
    
#ifdef DEBUG1
    printf ("Start = %d, Size = %d, start+size= %d, fsize = %d.\n", 
	    start, size, start+size, FILE_SIZE);
#endif
    lseek (holey_fd, start, SEEK_SET);
    buf_start = 0;
    write_size = size;

    /* Some file systems will only do partial writes.  According to POSIX,
       that's OK.  We loop here until we've written everything we want to.
    */
    while ( (nwritten = write (holey_fd, &buf[buf_start], write_size)) != write_size) {

      /* Oops, error condition. */
      if (nwritten == -1) {
	perror ("Error writing file.");
	exit (-1);
      }

      /* No error, just a short write. */
      printf ("Short write: write_size = %d, nwritten = %d\n", write_size, nwritten);
      buf_start += nwritten;
      write_size -= nwritten;
    }; 

#ifdef DEBUG1
    written += (long long) size;

    /* Short test that succeeds whenever we've written another 1MB (approx.)
     * so we can print a progress message.
     */

    newly_written += (long long) size;
    if (newly_written > 10000000) {
      printf ("%d bytes written.\n", written);
      newly_written = 0;
    }
#endif

    /* do the same copy in memory */
    memcpy (&holey_image[start], buf, size);

    /* Re-read the entire file into memory and them compare against our
       in-memory version */
    lseek (holey_fd, 0, SEEK_SET);
    if ( (read (holey_fd, holey_image_mem, FILE_SIZE)) != FILE_SIZE) {
      perror ("Rereading of file failed");
      exit (-1);
    }
   
    err = memcmp (holey_image, holey_image_mem, FILE_SIZE);
    /* Check to see if file == in-memory image */
    if (err) {
      printf ("Error: memcmp returned %d.\n", err);

      /* find the offset where they differ */
      j = 0;
      while (holey_image[j] == holey_image_mem[j]) 
	j++;

      printf ("File image is incorrect at offset %d.\n\n", j);
      printf ("Here are the context bytes:\n");
      printf ("Offset\tWritten\tRead\n");
      for (i = j - ERR_CONTEXT; i < j + ERR_CONTEXT; i++) 
	printf ("%6d\t0x%0x\t0x%0x\n", i, holey_image[i], holey_image_mem[i]);
      exit (-1);
    }

  } /* for (i... */

  exit (0);
} /* main */
    
  
