#!/bin/sh
#
# Stresses NFS reads and writes.
#
# Options: none
#
# Requires: cat, cp, diff, mkdir, mount, rm, rmdir,
#           init.d/nfs script
#

# cleanup is called right before we exit.  ERROR_VALUE tells us
# what we need to back out of.
#

# Restore nfs to it's state prior to execution of this script
restore_nfs_state () {
    if [ $NFS_STATUS = "reload" ] ; then # NFS was running
	/etc/rc.d/init.d/nfs reload >> /dev/null 2>&1
    elif [ $NFS_STATUS = "start" ] ; then # NFS was not running
	/etc/rc.d/init.d/nfs stop >> /dev/null 2>&1
    else                                  # Can't happen
       echo "Can't determine original NFS state."
    fi
}

# Restore /etc/exports to its original state, if it existed.  
# If not, delete it.
restore_etc_exports () {
    if [ $RESTORE_EXPORTS_FILE = "yes" ] ; then {
	rm -f $EXPORTS_FILE
	mv $EXPORTS_FILE_BACKUP $EXPORTS_FILE
    } ;
    elif [ $RESTORE_EXPORTS_FILE = "no" ] ; then
	rm -f $EXPORTS_FILE
    else # Can't happen
	echo "Can't determine original state of /etc/exports."
    fi
}

# Get rid of any temp dirs we may have created.  Not all of those might exist.
# This has to be done after TARGET0 has been unmounted, else we destroy
# /usr/src/linux.
remove_tmp_dirs () {    
    rmdir $TARGET0 >> /dev/null 2>&1
    if [ $? -ne 0 ] ; then echo "Can't rmdir $TARGET0." ; fi

    rm -r $SOURCE1 >> /dev/null 2>&1
    if [ $? -ne 0 ] ; then echo "Can't rm $SOURCE1." ; fi
    
    rm -r $TARGET1 >> /dev/null 2>&1
    if [ $? -ne 0 ] ; then echo "Can't rm $TARGET1." ; fi
}



# 
# Depending on how far through the script we got, we might have to 
# perform multiple steps to restore the previous state.  This function
# does the necessary steps
#
function cleanup () {
case "$ERROR_VALUE"
in
#
# 0 means success, just back out of all changes

    0 )	umount $TARGET1 >> /dev/null 2>&1
        if [ $? -ne 0 ] ; then {
	    echo "Can't umount $TARGET1."
	    exit 1
	} ; fi

	umount $TARGET0 >> /dev/null 2>&1
	if [ $? -ne 0 ] ; then {
	    echo "Can't umount $TARGET0."
	    exit 1
	} ; fi
        remove_tmp_dirs
	restore_etc_exports
	restore_nfs_state ;;

# 1,2 = error on startup, nothing to clean up.
    [12] ) ;;

# 3 = failed to make mount point directory
    3 ) restore_etc_exports ;;

# 4 = can't make temp dirs
# 5 = can't get probe NFS status
# 6 = bad NFS status
# 7 = can't modify NFS state
    [4567] ) restore_etc_exports
	    remove_tmp_dirs ;;

# 8 = can't mount on TARGET0
    8 ) restore_etc_exports
	restore_nfs_state
	remove_tmp_dirs ;;

# 9 = can't mount on TARGET1
    9 ) umount $TARGET0 >> /dev/null 2>&1
	if [ $? -ne 0 ] ; then {
	    echo "Can't umount $TARGET0."
	    echo "This is going to leave a mess."
	    exit 1
	} ; fi
	restore_etc_exports
	restore_nfs_state
	remove_tmp_dirs ;;

# 10 = error reading from NFS mounted dir
    10 ) umount $TARGET0 >> /dev/null 2>&1
	 if [ $? -ne 0 ] ; then {
	    echo "Can't umount $TARGET0."
	    echo "This is going to leave a mess."
	    exit 1
	 } ; fi
         umount $TARGET1 >> /dev/null 2>&1
	 if [ $? -ne 0 ] ; then {
	    echo "Can't umount $TARGET1."
	 } ; fi
	 restore_etc_exports
	 restore_nfs_state
	 remove_tmp_dirs ;;

# 11 = Can't write from client to server
# 12 = Error reading from client
    1[12] ) cd /usr/src # Let's not have cwd==$TARGET1 when we unmount it.
	    rm -rf $TARGET1/* >> /dev/null 2>&1
	    if [ $? -ne 0 ] ; then {
		echo "Can't rm -rf $TARGET1/*."
	    } ; fi
	    umount $TARGET0 >> /dev/null 2>&1
	    if [ $? -ne 0 ] ; then {
		echo "Can't umount $TARGET0."
		echo "This is going to leave a mess."
		exit 1
	    } ; fi
	    umount $TARGET1 >> /dev/null 2>&1
	    if [ $? -ne 0 ] ; then {
		echo "Can't umount $TARGET1."
	    } ; fi
	    remove_tmp_dirs 
	    restore_nfs_state 	
	    restore_etc_exports ;;

# 13 = Waiting for nfs to die.  Just exit.
    13 ) ;;

# anything else == bug in script
    * ) echo "Cleanup called with unknown error value.  This code is buggy." ;;

esac
}


# 
# Something's wrong, can't continue.
#
function bail () {
    echo $ERROR_MSG
    cleanup
    exit -2
}

#
# Find the name of a kernel tree in /usr/src.
#
function find_kernel_tree () {

[ -e /usr/src/cvs/linux -a \( -L /usr/src/cvs/linux -o -d /usr/src/cvs/linux \) ] \
    && { SOURCE0="/usr/src/cvs/linux" ; return 0 ; }
[ -e /usr/src/linux-2.4 -a \( -L /usr/src/linux-2.4 -o -d /usr/src/linux-2.4 \) ] \
    && { SOURCE0="/usr/src/linux-2.4" ; return 0 ; }
[ -e /usr/src/linux -a \( -L /usr/src/linux -o -d /usr/src/linux \) ] \
    && { SOURCE0="/usr/src/linux" ; return 0 ; }
return 1
}
    
#
# This is the beginning of the NFS test.
#
# For the first test, we are going to nfs mount the local machine's
# kernel tree on the local machine.
# We mount it on the same machine it's exported from to avoid going too
# deep into the IP stack; in particular, we don't want to dip into the
# NIC layer.
#

# find_kernel_tree sets SOURCE0 to kernel tree, or returns 1 if it can't find one.
find_kernel_tree
[ $? -eq 0 ] || { echo "Can't find a kernel tree in /usr/src." ; exit 1 ; }


SOURCE1="/usr/src/linux.tmp.$$-s1"
TARGET0="/usr/src/linux.tmp.$$-t0"
TARGET1="/usr/src/linux.tmp.$$-t1"

HOST=127.0.0.1
EXIT_VALUE=0
ERROR_MSG=""

#
# Def's for config files
# 
EXPORTS_FILE="/etc/exports"
EXPORTS_FILE_BACKUP="/etc/exports.original.$$"

#
# Let's make sure $SOURCE0 is a legitimate source.
#
if [  ! -e $SOURCE0 ] ; then {
    ERROR_MSG="$SOURCE0 does not exist."
    ERROR_VALUE=1
    bail
} ; fi

if [ ! -d $SOURCE0 ] ; then {
    ERROR_MSG="$SOURCE0 is not a directory\!."
    ERROR_VALUE=2
    bail
} ; fi

#
# Modify exports file
#
if [ ! -f $EXPORTS_FILE ] ; then {
    RESTORE_EXPORTS_FILE="no"
    cat > $EXPORTS_FILE <<EOF
# /etc/exports
#
# Async is the default behaviour.  Does this help stress testing?
#
$SOURCE0	$HOST(rw,no_root_squash,async)
$SOURCE1        $HOST(rw,no_root_squash,async)
EOF
} ; 

else {
    RESTORE_EXPORTS_FILE="yes"
    cp $EXPORTS_FILE $EXPORTS_FILE_BACKUP >>/dev/null 2>&1
    if [ $? -ne 0 ] ; then {
	ERROR_MSG="Can't backup $EXPORTS_FILE."
	ERROR_VALUE=2
	bail
    } ; fi
    cat >> $EXPORTS_FILE <<EOF
#
# Async is the default behaviour.  Does this help stress testing?
#
$SOURCE0	$HOST(rw,no_root_squash,async)
$SOURCE1        $HOST(rw,no_root_squash,async)
EOF
} ; fi

trap "ERROR_VALUE=3; bail" SIGHUP SIGINT SIGTERM

#
# Setup the directories for mounting, exporting
#
mkdir $SOURCE1 >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't make directory $SOURCE1."
    ERROR_VAL=4
    bail
} ; fi

mkdir $TARGET0 >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't make directory $TARGET0."
    ERROR_VAL=4
    bail
} ; fi

mkdir $TARGET1 >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't make directory $TARGET1."
    ERROR_VAL=4
    bail
} ; fi

trap "ERROR_VALUE=4; bail" SIGHUP SIGINT SIGTERM

#
# Check nfs status, start if necessary
#
NFS_STATUS=`/etc/rc.d/init.d/nfs probe` >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't probe NFS status."
    ERROR_VALUE=5
    bail
} ; fi

#
# $NFS_STATUS should be "start" or "reload" here.  If not, there's a bug
# in this script.
#
if [ ! "$NFS_STATUS" ] ; then {
    ERROR_MSG="Bad NFS status NULL."
    ERROR_VALUE=6
    bail
} ; fi
if [ "$NFS_STATUS" != "start" -a "$NFS_STATUS" != "reload" ] ; then {
    ERROR_MSG="Bad NFS status $NFS_STATUS."
    ERROR_VALUE=6
    bail
} ; fi

/etc/rc.d/init.d/nfs $NFS_STATUS >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't modify NFS status using \
	    \"/etc/rc.d/init.d/nfs $NFS_STATUS\"."
    ERROR_VALUE=7
    bail
} ; fi

trap "ERROR_VALUE=8; bail" SIGHUP SIGINT SIGTERM

#
# Mount the NFS exported directories
#
mount -t nfs "$HOST":"$SOURCE0" "$TARGET0" >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't mount $HOST:$SOURCE0 on $TARGET0."
    ERROR_VALUE=8
    bail
} ; fi
mount -t nfs "$HOST":"$SOURCE1" "$TARGET1" >> /dev/null 2>&1
if [ $? -ne 0 ] ; then {
    ERROR_MSG="Can't mount $HOST:$SOURCE1 on $TARGET1."
    ERROR_VALUE=9
    bail
} ; fi
#
# Blechh.  Broken symlinks causes diff to consider two identical directories
# to be different.  But, it returns 2 (rather than 1) if broken
# symlinks are the only problem (I think.)
#
# Also, maybe checksumming would be a more intelligent way of checking to
# make sure NFS read is idempotent?
#

trap "ERROR_VALUE=10; bail" SIGHUP SIGINT SIGTERM

# 
# SOURCE0 is /usr/src/linux, TARGET0 is the client mounted version
# of SOURCE0

diff -qr $SOURCE0 $TARGET0 >> /dev/null 2>&1
EXIT_VALUE=$?

if [ $? -eq 2 ] ; then {
    echo "You might have broken symlinks in $SOURCE0." 
#    EXIT_VALUE=2
# Do we want to return an error on this?
#
} ; fi

if [ $? -eq 1 ] ; then {
    ERROR_MSG="There is an error reading from an NFS mounted directory."
    ERROR_VALUE=10
    bail
} ; fi

#
# Push to server from client.
#
# First, populate TARGET1, the second client directory from SOURCE0,
# which is /usr/src/linux.
#

# 
# We need -d, otherwise cp will bail on broken symlinks.
# Should also check just to make sure /usr/src has enough free space.
#

trap "ERROR_VALUE=12; bail" SIGINT SIGHUP SIGTERM

for I in $SOURCE0/* $SOURCE0/.* ; do {
    if [ $I != "/usr/src/linux/." -a $I != "/usr/src/linux/.." \
	-a $I != "/usr/src/linux-2.4/." -a $I != "/usr/src/linux-2.4/.." \
        -a $I != "/usr/src/cvs/linux/." -a $I != "/usr/src/cvs/linux/.." ] ; then {
	cp -pRd "$I" "$TARGET1" >> /dev/null 2>&1
	if [ $? -ne 0 ] ; then {
	    ERROR_MSG="Copy to server failed."
	    ERROR_VALUE=11
	    bail
	} ; fi
    } ; fi
} ; done

diff -qr $SOURCE0 $TARGET1 >> /dev/null 2>&1
EXIT_VALUE=$?

if [ $? -eq 1 ] ; then {
    ERROR_MSG="Error writing to an NFS mounted directory."
    ERROR_VALUE=12
    bail
} ; fi

HOME_DIR=`pwd`
trap "cd $HOME_DIR ; ERROR_VALUE=12 ; bail" SIGHUP SIGINT SIGTERM

cd "$TARGET1"
[ $? -eq 0 ] || { ERROR_MSG="Error changing to dir $TARGET1." ; ERROR_VALUE=12 ; bail ; }

make mrproper
[ $? -eq 0 ] || { ERROR_MSG="Make mrproper failed." ; ERROR_VALUE=12 ; cd $HOME_DIR ; bail ; }

yes "" | make oldconfig
 [ $? -eq 0 ] || { ERROR_MSG="Make oldconfig failed." ; ERROR_VALUE=12 ; cd $HOME_DIR ; bail ; }

make -j 3 dep 
[ $? -eq 0 ] || { ERROR_MSG="Make dep failed." ; ERROR_VALUE=12 ; cd $HOME_DIR ; bail ; }

make clean
[ $? -eq 0 ] || { ERROR_MSG="Make clean failed." ; ERROR_VALUE=12 ; cd $HOME_DIR ; bail ; }

#
# We "make vmlinux" here, because it exists on all platforms.  Not
# so for bzImage
#
make -j 3 vmlinux
[ $? -eq 0 ] || { ERROR_MSG="Make vmlinux failed." ; ERROR_VALUE=12 ; cd $HOME_DIR ; bail ; }

make -j 3 modules
[ $? -eq 0 ] || { ERROR_MSG="Make modules failed." ; ERROR_VALUE=12 ; cd $HOME_DIR ; bail ; }

cd $HOME_DIR

ERROR_VALUE=12
cleanup

#
# We've been seeing some errors related to a race condition between the
# script and nfsd.
# If we try to re-execute the script immediatly after it's termination, we
# might try to start nfs services before all the old "nfsd" processes have
# died.  This causes subsequent nfs mounts to hang.
#
# Check NFS status.  If it was not running prior to the test, wait
# for nfsd's to die. 
#

if [ $NFS_STATUS = "start" ] ; then { #NFS was not running

    trap "cd $HOME_DIR ; ERROR_VALUE=13 ; exit 0" SIGHUP SIGINT SIGTERM
    NFS_DEAD=0
    while [ $NFS_DEAD -eq 0 ] ; do {
	ps -aux | grep 'nfsd' | grep -v 'grep' >> /dev/null 2>&1
	NFS_DEAD=$?
	sleep 10
    } ; done

} ; fi

exit 0

